@@ -23,14 +23,17 @@
 
 BSON_BEGIN_DECLS
 
+#define mcommon_md5_init COMMON_NAME (md5_init)
+#define mcommon_md5_append COMMON_NAME (md5_append)
+#define mcommon_md5_finish COMMON_NAME (md5_finish)
+
 void
-_bson_md5_init (bson_md5_t *pms);
+mcommon_md5_init (bson_md5_t *pms);
 void
-_bson_md5_append (bson_md5_t *pms, const uint8_t *data, uint32_t nbytes);
+mcommon_md5_append (bson_md5_t *pms, const uint8_t *data, uint32_t nbytes);
 void
-_bson_md5_finish (bson_md5_t *pms, uint8_t digest[16]);
+mcommon_md5_finish (bson_md5_t *pms, uint8_t digest[16]);
 
 BSON_END_DECLS
 
 #endif /* COMMON_MD5_PRIVATE_H */
-
