@@ -18,7 +18,7 @@
 
 #ifdef MONGOC_ENABLE_SSL_OPENSSL
 
-#include <bson.h>
+#include <bson/bson.h>
 
 #include <errno.h>
 #include <string.h>
@@ -29,6 +29,8 @@
 
 #include "mongoc-counters-private.h"
 #include "mongoc-errno-private.h"
+#include "mongoc-ssl.h"
+#include "mongoc-ssl-private.h"
 #include "mongoc-stream-tls.h"
 #include "mongoc-stream-private.h"
 #include "mongoc-stream-tls-private.h"
@@ -45,7 +47,8 @@
 
 #define MONGOC_STREAM_TLS_OPENSSL_BUFFER_SIZE 4096
 
-#if OPENSSL_VERSION_NUMBER < 0x10100000L || (defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x2070000fL)
+#if OPENSSL_VERSION_NUMBER < 0x10100000L || \
+   (defined(LIBRESSL_VERSION_NUMBER) && LIBRESSL_VERSION_NUMBER < 0x20700000L)
 static void
 BIO_meth_free (BIO_METHOD *meth)
 {
@@ -93,6 +96,9 @@ _mongoc_stream_tls_openssl_destroy (mongoc_stream_t *stream)
    SSL_CTX_free (openssl->ctx);
    openssl->ctx = NULL;
 
+   mongoc_openssl_ocsp_opt_destroy (openssl->ocsp_opts);
+   openssl->ocsp_opts = NULL;
+
    bson_free (openssl);
    bson_free (stream);
 
@@ -537,14 +543,38 @@ _mongoc_stream_tls_openssl_check_closed (mongoc_stream_t *stream) /* IN */
 }
 
 
+static bool
+_mongoc_stream_tls_openssl_set_verify_cert_error (SSL *ssl, bson_error_t *error)
+{
+   long verify_result;
+
+   BSON_ASSERT_PARAM (ssl);
+   BSON_ASSERT_PARAM (error);
+
+   verify_result = SSL_get_verify_result (ssl);
+
+   if (verify_result == X509_V_OK) {
+      return false;
+   }
+
+   bson_set_error (error,
+                   MONGOC_ERROR_STREAM,
+                   MONGOC_ERROR_STREAM_SOCKET,
+                   "TLS handshake failed: certificate verify failed (%ld): %s",
+                   verify_result,
+                   X509_verify_cert_error_string (verify_result));
+
+   return true;
+}
+
 /**
  * mongoc_stream_tls_openssl_handshake:
  */
-bool
-mongoc_stream_tls_openssl_handshake (mongoc_stream_t *stream,
-                                     const char *host,
-                                     int *events,
-                                     bson_error_t *error)
+static bool
+_mongoc_stream_tls_openssl_handshake (mongoc_stream_t *stream,
+                                      const char *host,
+                                      int *events,
+                                      bson_error_t *error)
 {
    mongoc_stream_tls_t *tls = (mongoc_stream_tls_t *) stream;
    mongoc_stream_tls_openssl_t *openssl =
@@ -558,12 +588,31 @@ mongoc_stream_tls_openssl_handshake (mongoc_stream_t *stream,
    BIO_get_ssl (openssl->bio, &ssl);
 
    if (BIO_do_handshake (openssl->bio) == 1) {
-      if (_mongoc_openssl_check_cert (
+      *events = 0;
+
+#ifdef MONGOC_ENABLE_OCSP_OPENSSL
+      /* Validate OCSP */
+      if (openssl->ocsp_opts &&
+          1 != _mongoc_ocsp_tlsext_status (ssl, openssl->ocsp_opts)) {
+         bson_set_error (error,
+                         MONGOC_ERROR_STREAM,
+                         MONGOC_ERROR_STREAM_SOCKET,
+                         "TLS handshake failed: Failed OCSP verification");
+         RETURN (false);
+      }
+#endif
+
+      if (_mongoc_openssl_check_peer_hostname (
              ssl, host, tls->ssl_opts.allow_invalid_hostname)) {
          RETURN (true);
       }
 
-      *events = 0;
+      /* Try to relay certificate failure reason from OpenSSL library if any. */
+      if (_mongoc_stream_tls_openssl_set_verify_cert_error (ssl, error)) {
+         RETURN (false);
+      }
+
+      /* Otherwise, use simple error message. */
       bson_set_error (error,
                       MONGOC_ERROR_STREAM,
                       MONGOC_ERROR_STREAM_SOCKET,
@@ -585,13 +634,50 @@ mongoc_stream_tls_openssl_handshake (mongoc_stream_t *stream,
 #endif
    }
 
-
    *events = 0;
-   bson_set_error (error,
-                   MONGOC_ERROR_STREAM,
-                   MONGOC_ERROR_STREAM_SOCKET,
-                   "TLS handshake failed: %s",
-                   ERR_error_string (ERR_get_error (), NULL));
+
+   /* Try to relay certificate failure reason from OpenSSL library if any. */
+   if (_mongoc_stream_tls_openssl_set_verify_cert_error (ssl, error)) {
+      RETURN (false);
+   }
+
+   /* Otherwise, try to relay error info from OpenSSL. */
+   if (ERR_peek_error () != 0) {
+      bson_set_error (error,
+                      MONGOC_ERROR_STREAM,
+                      MONGOC_ERROR_STREAM_SOCKET,
+                      "TLS handshake failed: %s",
+                      ERR_error_string (ERR_get_error (), NULL));
+      RETURN (false);
+   }
+
+   /* Otherwise, use simple error info. */
+   {
+#ifdef _WIN32
+      LPTSTR msg = NULL;
+      FormatMessage (FORMAT_MESSAGE_ALLOCATE_BUFFER |
+                        FORMAT_MESSAGE_FROM_SYSTEM |
+                        FORMAT_MESSAGE_ARGUMENT_ARRAY,
+                     NULL,
+                     errno, /* WSAETIMEDOUT */
+                     LANG_NEUTRAL,
+                     (LPTSTR) &msg,
+                     0,
+                     NULL);
+#else
+      const char *msg = strerror (errno); /* ETIMEDOUT */
+#endif
+
+      bson_set_error (error,
+                      MONGOC_ERROR_STREAM,
+                      MONGOC_ERROR_STREAM_SOCKET,
+                      "TLS handshake failed: %s",
+                      msg);
+
+#ifdef _WIN32
+      LocalFree (msg);
+#endif
+   }
 
    RETURN (false);
 }
@@ -605,6 +691,9 @@ _mongoc_stream_tls_openssl_sni (SSL *ssl, int *ad, void *arg)
 {
    const char *hostname;
 
+   BSON_UNUSED (ad);
+   BSON_UNUSED (arg);
+
    if (ssl == NULL) {
       TRACE ("%s", "No SNI hostname provided");
       return SSL_TLSEXT_ERR_NOACK;
@@ -627,6 +716,22 @@ _mongoc_stream_tls_openssl_timed_out (mongoc_stream_t *stream)
    RETURN (mongoc_stream_timed_out (tls->base_stream));
 }
 
+static bool
+_mongoc_stream_tls_openssl_should_retry (mongoc_stream_t *stream)
+{
+   mongoc_stream_tls_t *tls = (mongoc_stream_tls_t *) stream;
+   mongoc_stream_tls_openssl_t *openssl =
+      (mongoc_stream_tls_openssl_t *) tls->ctx;
+
+   ENTRY;
+
+   if (BIO_should_retry (openssl->bio)) {
+      RETURN (true);
+   }
+
+   RETURN (mongoc_stream_should_retry (tls->base_stream));
+}
+
 /*
  *--------------------------------------------------------------------------
  *
@@ -658,6 +763,7 @@ mongoc_stream_tls_openssl_new (mongoc_stream_t *base_stream,
 {
    mongoc_stream_tls_t *tls;
    mongoc_stream_tls_openssl_t *openssl;
+   mongoc_openssl_ocsp_opt_t *ocsp_opts = NULL;
    SSL_CTX *ssl_ctx = NULL;
    BIO *bio_ssl = NULL;
    BIO *bio_mongoc_shim = NULL;
@@ -676,12 +782,13 @@ mongoc_stream_tls_openssl_new (mongoc_stream_t *base_stream,
 #if OPENSSL_VERSION_NUMBER >= 0x10002000L && !defined(LIBRESSL_VERSION_NUMBER)
    if (!opt->allow_invalid_hostname) {
       struct in_addr addr;
+      struct in6_addr addr6;
       X509_VERIFY_PARAM *param = X509_VERIFY_PARAM_new ();
 
       X509_VERIFY_PARAM_set_hostflags (param,
                                        X509_CHECK_FLAG_NO_PARTIAL_WILDCARDS);
       if (inet_pton (AF_INET, host, &addr) ||
-          inet_pton (AF_INET6, host, &addr)) {
+          inet_pton (AF_INET6, host, &addr6)) {
          X509_VERIFY_PARAM_set1_ip_asc (param, host);
       } else {
          X509_VERIFY_PARAM_set1_host (param, host, 0);
@@ -714,6 +821,7 @@ mongoc_stream_tls_openssl_new (mongoc_stream_t *base_stream,
    if (!bio_mongoc_shim) {
       BIO_free_all (bio_ssl);
       BIO_meth_free (meth);
+      SSL_CTX_free (ssl_ctx);
       RETURN (NULL);
    }
 
@@ -727,13 +835,43 @@ mongoc_stream_tls_openssl_new (mongoc_stream_t *base_stream,
 #endif
    }
 
-
    BIO_push (bio_ssl, bio_mongoc_shim);
 
+#ifdef MONGOC_ENABLE_OCSP_OPENSSL
+   if (client && !opt->weak_cert_validation &&
+       !_mongoc_ssl_opts_disable_certificate_revocation_check (opt)) {
+      SSL *ssl;
+
+      BIO_get_ssl (bio_ssl, &ssl);
+
+      /* Set the status_request extension on the SSL object.
+       * Do not use SSL_CTX_set_tlsext_status_type, since that requires OpenSSL
+       * 1.1.0.
+       */
+      if (!SSL_set_tlsext_status_type (ssl, TLSEXT_STATUSTYPE_ocsp)) {
+         MONGOC_ERROR ("cannot enable OCSP status request extension");
+         mongoc_openssl_ocsp_opt_destroy (ocsp_opts);
+         BIO_free_all (bio_ssl);
+         BIO_meth_free (meth);
+         SSL_CTX_free (ssl_ctx);
+         RETURN (NULL);
+      }
+
+      ocsp_opts = bson_malloc0 (sizeof (mongoc_openssl_ocsp_opt_t));
+      ocsp_opts->allow_invalid_hostname = opt->allow_invalid_hostname;
+      ocsp_opts->weak_cert_validation = opt->weak_cert_validation;
+      ocsp_opts->disable_endpoint_check =
+         _mongoc_ssl_opts_disable_ocsp_endpoint_check (opt);
+      ocsp_opts->host = bson_strdup (host);
+      _mongoc_ssl_opts_copy_to (opt, &ocsp_opts->ssl_opts, true);
+   }
+#endif /* MONGOC_ENABLE_OCSP_OPENSSL */
+
    openssl = (mongoc_stream_tls_openssl_t *) bson_malloc0 (sizeof *openssl);
    openssl->bio = bio_ssl;
    openssl->meth = meth;
    openssl->ctx = ssl_ctx;
+   openssl->ocsp_opts = ocsp_opts;
 
    tls = (mongoc_stream_tls_t *) bson_malloc0 (sizeof *tls);
    tls->parent.type = MONGOC_STREAM_TLS;
@@ -747,8 +885,9 @@ mongoc_stream_tls_openssl_new (mongoc_stream_t *base_stream,
    tls->parent.get_base_stream = _mongoc_stream_tls_openssl_get_base_stream;
    tls->parent.check_closed = _mongoc_stream_tls_openssl_check_closed;
    tls->parent.timed_out = _mongoc_stream_tls_openssl_timed_out;
+   tls->parent.should_retry = _mongoc_stream_tls_openssl_should_retry;
    memcpy (&tls->ssl_opts, opt, sizeof tls->ssl_opts);
-   tls->handshake = mongoc_stream_tls_openssl_handshake;
+   tls->handshake = _mongoc_stream_tls_openssl_handshake;
    tls->ctx = (void *) openssl;
    tls->timeout_msec = -1;
    tls->base_stream = base_stream;
@@ -759,4 +898,18 @@ mongoc_stream_tls_openssl_new (mongoc_stream_t *base_stream,
    RETURN ((mongoc_stream_t *) tls);
 }
 
+void
+mongoc_openssl_ocsp_opt_destroy (void *ocsp_opt)
+{
+   mongoc_openssl_ocsp_opt_t *casted;
+
+   if (!ocsp_opt) {
+      return;
+   }
+   casted = (mongoc_openssl_ocsp_opt_t *) ocsp_opt;
+   bson_free (casted->host);
+   _mongoc_ssl_opts_cleanup (&casted->ssl_opts, true);
+   bson_free (ocsp_opt);
+}
+
 #endif /* MONGOC_ENABLE_SSL_OPENSSL */
