@@ -53,7 +53,7 @@
 
 #ifdef MONGOC_ENABLE_SSL_SECURE_CHANNEL
 
-#include <bson.h>
+#include <bson/bson.h>
 
 #include "mongoc-trace-private.h"
 #include "mongoc-log.h"
@@ -63,6 +63,7 @@
 #include "mongoc-stream-tls-secure-channel-private.h"
 #include "mongoc-secure-channel-private.h"
 #include "mongoc-ssl.h"
+#include "mongoc-ssl-private.h"
 #include "mongoc-error.h"
 #include "mongoc-counters-private.h"
 #include "mongoc-errno-private.h"
@@ -106,7 +107,7 @@ _mongoc_stream_tls_secure_channel_destroy (mongoc_stream_t *stream)
     * Shutting Down an Schannel Connection
     */
 
-   TRACE ("shutting down SSL/TLS connection");
+   TRACE ("%s", "shutting down SSL/TLS connection");
 
    if (secure_channel->cred && secure_channel->ctxt) {
       SecBufferDesc BuffDesc;
@@ -124,7 +125,7 @@ _mongoc_stream_tls_secure_channel_destroy (mongoc_stream_t *stream)
          ApplyControlToken (&secure_channel->ctxt->ctxt_handle, &BuffDesc);
 
       if (sspi_status != SEC_E_OK) {
-         MONGOC_ERROR ("ApplyControlToken failure: %d", sspi_status);
+         MONGOC_ERROR ("ApplyControlToken failure: %ld", sspi_status);
       }
 
       /* setup output buffer */
@@ -154,7 +155,7 @@ _mongoc_stream_tls_secure_channel_destroy (mongoc_stream_t *stream)
          FreeContextBuffer (outbuf.pvBuffer);
 
          if (outbuf.cbBuffer != (size_t) written) {
-            TRACE ("failed to send close msg (wrote %zd out of %zd)",
+            TRACE ("failed to send close msg (wrote %zd out of %lu)",
                    written,
                    outbuf.cbBuffer);
          }
@@ -163,7 +164,7 @@ _mongoc_stream_tls_secure_channel_destroy (mongoc_stream_t *stream)
 
    /* free SSPI Schannel API security context handle */
    if (secure_channel->ctxt) {
-      TRACE ("clear security context handle");
+      TRACE ("%s", "clear security context handle");
       DeleteSecurityContext (&secure_channel->ctxt->ctxt_handle);
       bson_free (secure_channel->ctxt);
    }
@@ -172,7 +173,7 @@ _mongoc_stream_tls_secure_channel_destroy (mongoc_stream_t *stream)
    if (secure_channel->cred) {
       /* decrement the reference counter of the credential/session handle */
       /* if the handle was not cached and the refcount is zero */
-      TRACE ("clear credential handle");
+      TRACE ("%s", "clear credential handle");
       FreeCredentialsHandle (&secure_channel->cred->cred_handle);
       bson_free (secure_channel->cred);
    }
@@ -254,7 +255,7 @@ _mongoc_stream_tls_secure_channel_write (mongoc_stream_t *stream,
    ENTRY;
 
    BSON_ASSERT (secure_channel);
-   TRACE ("The entire buffer is: %d", buf_len);
+   TRACE ("The entire buffer is: %zu", buf_len);
 
    /* check if the maximum stream sizes were queried */
    if (secure_channel->stream_sizes.cbMaximumMessage == 0) {
@@ -270,7 +271,7 @@ _mongoc_stream_tls_secure_channel_write (mongoc_stream_t *stream,
 
    /* check if the buffer is longer than the maximum message length */
    if (buf_len > secure_channel->stream_sizes.cbMaximumMessage) {
-      TRACE ("SHRINKING buf_len from %lu to %lu",
+      TRACE ("SHRINKING buf_len from %zu to %lu",
              buf_len,
              secure_channel->stream_sizes.cbMaximumMessage);
       buf_len = secure_channel->stream_sizes.cbMaximumMessage;
@@ -370,15 +371,15 @@ _mongoc_stream_tls_secure_channel_writev (mongoc_stream_t *stream,
    BSON_ASSERT (secure_channel);
    ENTRY;
 
-   TRACE ("Trying to write to the server");
+   TRACE ("%s", "Trying to write to the server");
    tls->timeout_msec = timeout_msec;
 
-   TRACE ("count: %d, 0th: %lu", iovcnt, iov[0].iov_len);
+   TRACE ("count: %zu, 0th: %zu", iovcnt, iov[0].iov_len);
 
    for (i = 0; i < iovcnt; i++) {
       iov_pos = 0;
 
-      TRACE ("iov %d size: %lu", i, iov[i].iov_len);
+      TRACE ("iov %zu size: %zu", i, iov[i].iov_len);
       while (iov_pos < iov[i].iov_len) {
          if (buf_head != buf_tail ||
              ((i + 1 < iovcnt) &&
@@ -419,7 +420,7 @@ _mongoc_stream_tls_secure_channel_writev (mongoc_stream_t *stream,
 
             child_ret = _mongoc_stream_tls_secure_channel_write (
                stream, to_write, to_write_len);
-            TRACE ("Child0wrote: %d, was supposed to write: %d",
+            TRACE ("Child0wrote: %zd, was supposed to write: %zu",
                    child_ret,
                    to_write_len);
 
@@ -441,7 +442,7 @@ _mongoc_stream_tls_secure_channel_writev (mongoc_stream_t *stream,
 
       child_ret = _mongoc_stream_tls_secure_channel_write (
          stream, buf_head, buf_tail - buf_head);
-      TRACE ("Child1wrote: %d, was supposed to write: %d",
+      TRACE ("Child1wrote: %zd, was supposed to write: %td",
              child_ret,
              buf_tail - buf_head);
 
@@ -456,7 +457,7 @@ _mongoc_stream_tls_secure_channel_writev (mongoc_stream_t *stream,
       mongoc_counter_streams_egress_add (ret);
    }
 
-   TRACE ("Returning %zu", ret);
+   TRACE ("Returning %d", (int) ret);
    RETURN (ret);
 }
 
@@ -474,10 +475,10 @@ _mongoc_stream_tls_secure_channel_debuf (
 
    secure_channel->decdata_offset -= s;
 
-   TRACE ("decrypted data returned %zu", s);
-   TRACE ("decrypted data buffer: offset %zu length %zu",
-          secure_channel->decdata_offset,
-          secure_channel->decdata_length);
+   TRACE ("decrypted data returned %d", (int) s);
+   TRACE ("decrypted data buffer: offset %d length %d",
+          (int) secure_channel->decdata_offset,
+          (int) secure_channel->decdata_length);
 
    return (ssize_t) s;
 }
@@ -490,15 +491,13 @@ _mongoc_stream_tls_secure_channel_decrypt (
 {
    size_t size = 0;
    size_t remaining;
-   ssize_t nread = -1;
-   bool done = false;
    SecBuffer inbuf[4];
    SecBufferDesc inbuf_desc;
    SECURITY_STATUS sspi_status = SEC_E_OK;
 
-   TRACE ("encrypted data buffer: offset %zu length %zu",
-          secure_channel->encdata_offset,
-          secure_channel->encdata_length);
+   TRACE ("encrypted data buffer: offset %d length %d",
+          (int) secure_channel->encdata_offset,
+          (int) secure_channel->encdata_length);
 
    /* decrypt loop */
    while (secure_channel->encdata_offset > 0 && sspi_status == SEC_E_OK) {
@@ -553,10 +552,10 @@ _mongoc_stream_tls_secure_channel_decrypt (
                secure_channel->decdata_offset += size;
             }
 
-            TRACE ("decrypted data added: %zu", size);
-            TRACE ("decrypted data cached: offset %zu length %zu",
-                   secure_channel->decdata_offset,
-                   secure_channel->decdata_length);
+            TRACE ("decrypted data added: %d", (int) size);
+            TRACE ("decrypted data cached: offset %d length %d",
+                   (int) secure_channel->decdata_offset,
+                   (int) secure_channel->decdata_length);
          }
 
          /* check for remaining encrypted data */
@@ -577,9 +576,9 @@ _mongoc_stream_tls_secure_channel_decrypt (
                secure_channel->encdata_offset = inbuf[3].cbBuffer;
             }
 
-            TRACE ("encrypted data cached: offset %zu length %zu",
-                   secure_channel->encdata_offset,
-                   secure_channel->encdata_length);
+            TRACE ("encrypted data cached: offset %d length %d",
+                   (int) secure_channel->encdata_offset,
+                   (int) secure_channel->encdata_length);
          } else {
             /* reset encrypted buffer offset, because there is no data remaining
              */
@@ -588,7 +587,7 @@ _mongoc_stream_tls_secure_channel_decrypt (
 
          /* check if server wants to renegotiate the connection context */
          if (sspi_status == SEC_I_RENEGOTIATE) {
-            TRACE ("remote party requests renegotiation");
+            TRACE ("%s", "remote party requests renegotiation");
          }
          /* check if the server closed the connection */
          else if (sspi_status == SEC_I_CONTEXT_EXPIRED) {
@@ -598,24 +597,24 @@ _mongoc_stream_tls_secure_channel_decrypt (
 
             if (!secure_channel->recv_connection_closed) {
                secure_channel->recv_connection_closed = true;
-               TRACE ("server closed the connection");
+               TRACE ("%s", "server closed the connection");
             }
          }
       } else if (sspi_status == SEC_E_INCOMPLETE_MESSAGE) {
-         TRACE ("failed to decrypt data, need more data");
+         TRACE ("%s", "failed to decrypt data, need more data");
       } else {
-         TRACE ("failed to read data from server: %d", sspi_status);
+         TRACE ("failed to read data from server: %ld", sspi_status);
          secure_channel->recv_unrecoverable_err = true;
       }
    }
 
-   TRACE ("encrypted data buffer: offset %zu length %zu",
-          secure_channel->encdata_offset,
-          secure_channel->encdata_length);
+   TRACE ("encrypted data buffer: offset %d length %d",
+          (int) secure_channel->encdata_offset,
+          (int) secure_channel->encdata_length);
 
-   TRACE ("decrypted data buffer: offset %zu length %zu",
-          secure_channel->decdata_offset,
-          secure_channel->decdata_length);
+   TRACE ("decrypted data buffer: offset %d length %d",
+          (int) secure_channel->decdata_offset,
+          (int) secure_channel->decdata_length);
 }
 
 
@@ -630,7 +629,7 @@ _mongoc_stream_tls_secure_channel_read (mongoc_stream_t *stream,
    ssize_t size = 0;
    ssize_t nread;
 
-   TRACE ("client wants to read %zu bytes", len);
+   TRACE ("client wants to read %d bytes", (int) len);
    BSON_ASSERT (len > 0);
 
    /*
@@ -640,7 +639,7 @@ _mongoc_stream_tls_secure_channel_read (mongoc_stream_t *stream,
     */
 
    if (secure_channel->decdata_offset) {
-      TRACE ("decrypted data is already available");
+      TRACE ("%s", "decrypted data is already available");
       return _mongoc_stream_tls_secure_channel_debuf (secure_channel, buf, len);
    }
 
@@ -655,17 +654,17 @@ _mongoc_stream_tls_secure_channel_read (mongoc_stream_t *stream,
 
    /* keep these checks separated, for more detailed tracing */
    if (secure_channel->recv_unrecoverable_err) {
-      TRACE ("an unrecoverable error occurred in a prior call");
+      TRACE ("%s", "an unrecoverable error occurred in a prior call");
       return -1;
    }
 
    if (secure_channel->recv_sspi_close_notify) {
-      TRACE ("server indicated shutdown in a prior call");
+      TRACE ("%s", "server indicated shutdown in a prior call");
       return -1;
    }
 
    if (secure_channel->recv_connection_closed) {
-      TRACE ("connection closed");
+      TRACE ("%s", "connection closed");
       return -1;
    }
 
@@ -680,7 +679,7 @@ _mongoc_stream_tls_secure_channel_read (mongoc_stream_t *stream,
 
    if (!nread) {
       if (MONGOC_ERRNO_IS_AGAIN (errno)) {
-         TRACE ("Try again");
+         TRACE ("%s", "Try again");
          return 0;
       } else {
          secure_channel->recv_connection_closed = true;
@@ -828,6 +827,10 @@ mongoc_stream_tls_secure_channel_handshake (mongoc_stream_t *stream,
    ENTRY;
    BSON_ASSERT (secure_channel);
 
+   if (error) {
+      error->code = 0;
+   }
+
    TRACE ("Getting ready for state: %d, timeout is %d",
           secure_channel->connecting_state + 1,
           tls->timeout_msec);
@@ -835,13 +838,12 @@ mongoc_stream_tls_secure_channel_handshake (mongoc_stream_t *stream,
    switch (secure_channel->connecting_state) {
    case ssl_connect_1:
 
-
-      if (mongoc_secure_channel_handshake_step_1 (tls, (char *) host)) {
-         TRACE ("Step#1 Worked!\n\n");
+      if (mongoc_secure_channel_handshake_step_1 (tls, (char *) host, error)) {
+         TRACE ("%s", "Step#1 Worked!\n\n");
          *events = POLLIN;
          RETURN (false);
       } else {
-         TRACE ("Step#1 FAILED!");
+         TRACE ("%s", "Step#1 FAILED!");
       }
 
       break;
@@ -850,7 +852,7 @@ mongoc_stream_tls_secure_channel_handshake (mongoc_stream_t *stream,
    case ssl_connect_2_reading:
    case ssl_connect_2_writing:
 
-      if (mongoc_secure_channel_handshake_step_2 (tls, (char *) host)) {
+      if (mongoc_secure_channel_handshake_step_2 (tls, (char *) host, error)) {
          if (secure_channel->connecting_state == ssl_connect_2_reading) {
             *events = POLLIN;
          } else {
@@ -858,37 +860,41 @@ mongoc_stream_tls_secure_channel_handshake (mongoc_stream_t *stream,
          }
          RETURN (false);
       } else {
-         TRACE ("Step#2 FAILED!");
+         TRACE ("%s", "Step#2 FAILED!");
       }
 
       break;
 
    case ssl_connect_3:
 
-      if (mongoc_secure_channel_handshake_step_3 (tls, (char *) host)) {
-         TRACE ("Step#3 Worked!\n\n");
+      if (mongoc_secure_channel_handshake_step_3 (tls, (char *) host, error)) {
+         TRACE ("%s", "Step#3 Worked!\n\n");
          *events = POLLIN | POLLOUT;
          RETURN (false);
       } else {
-         TRACE ("Step#3 FAILED!");
+         TRACE ("%s", "Step#3 FAILED!");
       }
 
       break;
 
    case ssl_connect_done:
-      TRACE ("Connect DONE!");
+      TRACE ("%s", "Connect DONE!");
       /* reset our connection state machine */
       secure_channel->connecting_state = ssl_connect_1;
       RETURN (true);
       break;
+   default: /* do nothing */
+      break;
    }
 
-
    *events = 0;
-   bson_set_error (error,
-                   MONGOC_ERROR_STREAM,
-                   MONGOC_ERROR_STREAM_SOCKET,
-                   "TLS handshake failed");
+
+   if (error && !error->code) {
+      bson_set_error (error,
+                      MONGOC_ERROR_STREAM,
+                      MONGOC_ERROR_STREAM_SOCKET,
+                      "TLS handshake failed");
+   }
 
    RETURN (false);
 }
@@ -903,6 +909,16 @@ _mongoc_stream_tls_secure_channel_timed_out (mongoc_stream_t *stream)
    RETURN (mongoc_stream_timed_out (tls->base_stream));
 }
 
+static bool
+_mongoc_stream_tls_secure_channel_should_retry (mongoc_stream_t *stream)
+{
+   mongoc_stream_tls_t *tls = (mongoc_stream_tls_t *) stream;
+
+   ENTRY;
+
+   RETURN (mongoc_stream_should_retry (tls->base_stream));
+}
+
 mongoc_stream_t *
 mongoc_stream_tls_secure_channel_new (mongoc_stream_t *base_stream,
                                       const char *host,
@@ -913,6 +929,7 @@ mongoc_stream_tls_secure_channel_new (mongoc_stream_t *base_stream,
    SCHANNEL_CRED schannel_cred;
    mongoc_stream_tls_t *tls;
    mongoc_stream_tls_secure_channel_t *secure_channel;
+   PCCERT_CONTEXT cert = NULL;
 
    ENTRY;
    BSON_ASSERT (base_stream);
@@ -942,13 +959,14 @@ mongoc_stream_tls_secure_channel_new (mongoc_stream_t *base_stream,
       _mongoc_stream_tls_secure_channel_get_base_stream;
    tls->parent.check_closed = _mongoc_stream_tls_secure_channel_check_closed;
    tls->parent.timed_out = _mongoc_stream_tls_secure_channel_timed_out;
+   tls->parent.should_retry = _mongoc_stream_tls_secure_channel_should_retry;
    memcpy (&tls->ssl_opts, opt, sizeof tls->ssl_opts);
    tls->handshake = mongoc_stream_tls_secure_channel_handshake;
    tls->ctx = (void *) secure_channel;
    tls->timeout_msec = -1;
    tls->base_stream = base_stream;
 
-   TRACE ("SSL/TLS connection with endpoint AcquireCredentialsHandle");
+   TRACE ("%s", "SSL/TLS connection with endpoint AcquireCredentialsHandle");
 
    /* setup Schannel API options */
    memset (&schannel_cred, 0, sizeof (schannel_cred));
@@ -960,20 +978,26 @@ mongoc_stream_tls_secure_channel_new (mongoc_stream_t *base_stream,
 #ifdef SCH_USE_STRONG_CRYPTO
    schannel_cred.dwFlags = SCH_USE_STRONG_CRYPTO;
 #endif
+
+   /* By default, enable soft failing.
+    * A certificate with no revocation check is a soft failure. */
+   schannel_cred.dwFlags |= SCH_CRED_IGNORE_NO_REVOCATION_CHECK;
+   /* An offline OCSP responder / CRL distribution list is a soft failure. */
+   schannel_cred.dwFlags |= SCH_CRED_IGNORE_REVOCATION_OFFLINE;
    if (opt->weak_cert_validation) {
-      schannel_cred.dwFlags |= SCH_CRED_MANUAL_CRED_VALIDATION |
-                               SCH_CRED_IGNORE_NO_REVOCATION_CHECK |
-                               SCH_CRED_IGNORE_REVOCATION_OFFLINE;
-      TRACE ("disabled server certificate checks");
+      schannel_cred.dwFlags |= SCH_CRED_MANUAL_CRED_VALIDATION;
+      TRACE ("%s", "disabled server certificate checks");
    } else {
-      schannel_cred.dwFlags |=
-         SCH_CRED_AUTO_CRED_VALIDATION | SCH_CRED_REVOCATION_CHECK_CHAIN;
-      TRACE ("enabled server certificate checks");
+      schannel_cred.dwFlags |= SCH_CRED_AUTO_CRED_VALIDATION;
+      if (!_mongoc_ssl_opts_disable_certificate_revocation_check (opt)) {
+         schannel_cred.dwFlags |= SCH_CRED_REVOCATION_CHECK_CHAIN;
+         TRACE ("%s", "enabled server certificate revocation checks");
+      }
+      TRACE ("%s", "enabled server certificate checks");
    }
 
    if (opt->allow_invalid_hostname) {
-      schannel_cred.dwFlags |=
-         SCH_CRED_NO_SERVERNAME_CHECK | SCH_CRED_IGNORE_NO_REVOCATION_CHECK;
+      schannel_cred.dwFlags |= SCH_CRED_NO_SERVERNAME_CHECK;
    }
 
    if (opt->ca_file) {
@@ -985,8 +1009,7 @@ mongoc_stream_tls_secure_channel_new (mongoc_stream_t *base_stream,
    }
 
    if (opt->pem_file) {
-      PCCERT_CONTEXT cert =
-         mongoc_secure_channel_setup_certificate (secure_channel, opt);
+      cert = mongoc_secure_channel_setup_certificate (secure_channel, opt);
 
       if (cert) {
          schannel_cred.cCreds = 1;
@@ -1030,8 +1053,8 @@ mongoc_stream_tls_secure_channel_new (mongoc_stream_t *base_stream,
                      NULL);
       MONGOC_ERROR (
          "Failed to initialize security context, error code: 0x%04X%04X: '%s'",
-         (sspi_status >> 16) & 0xffff,
-         sspi_status & 0xffff,
+         (unsigned int) (sspi_status >> 16) & 0xffff,
+         (unsigned int) sspi_status & 0xffff,
          msg);
       LocalFree (msg);
       RETURN (NULL);
@@ -1042,6 +1065,10 @@ mongoc_stream_tls_secure_channel_new (mongoc_stream_t *base_stream,
                     "against Secure Channel");
    }
 
+   if (_mongoc_ssl_opts_disable_ocsp_endpoint_check (opt)) {
+      MONGOC_ERROR ("Setting tlsDisableOCSPEndpointCheck has no effect when "
+                    "built against Secure Channel");
+   }
 
    mongoc_counter_streams_active_inc ();
    RETURN ((mongoc_stream_t *) tls);
